/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2019 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.archetype.rules.util;

import org.openvpms.component.business.domain.im.common.Entity;
import org.openvpms.component.business.domain.im.common.EntityRelationship;
import org.openvpms.component.business.domain.im.common.IMObjectReference;
import org.openvpms.component.business.service.archetype.ArchetypeServiceException;
import org.openvpms.component.business.service.archetype.IArchetypeService;
import org.openvpms.component.business.service.archetype.helper.IMObjectBeanException;
import org.openvpms.component.model.bean.IMObjectBean;
import org.openvpms.component.model.object.Reference;
import org.openvpms.component.model.object.Relationship;

import java.util.List;


/**
 * Entity relationship helper methods.
 *
 * @author Tim Anderson
 */
public class EntityRelationshipHelper {

    /**
     * Returns the target from the default entity relationship from the specified relationship node.
     * <p/>
     * Note that this can return a reference to an inactive object.
     *
     * @param entity  the parent entity
     * @param node    the relationship node
     * @param service the archetype service
     * @return the default target, or the the first target if there is no default, or {@code null} if none is found
     * @throws ArchetypeServiceException for any archetype service error
     * @throws IMObjectBeanException     if the node doesn't exist or an element is of the wrong type
     */
    public static IMObjectReference getDefaultTargetRef(
            Entity entity, String node, IArchetypeService service) {
        return getDefaultTargetRef(entity, node, true, service);
    }

    /**
     * Returns the target from the default entity relationship from the specified relationship node.
     * <p/>
     * Note that this can return a reference to an inactive object.
     *
     * @param entity        the parent entity
     * @param node          the relationship node
     * @param useNonDefault if {@code true} use a non-default relationship if no default can be found
     * @param service       the archetype service
     * @return the default target, or the the first target if there is no
     * default, or {@code null} if none is found
     * @throws ArchetypeServiceException for any archetype service error
     * @throws IMObjectBeanException     if the node doesn't exist or an element is of the wrong type
     */
    public static IMObjectReference getDefaultTargetRef(org.openvpms.component.model.entity.Entity entity,
                                                        String node, boolean useNonDefault,
                                                        IArchetypeService service) {
        IMObjectBean bean = service.getBean(entity);
        return getDefaultTargetRef(bean.getValues(node, EntityRelationship.class), useNonDefault, service);
    }

    /**
     * Returns the active target from the default entity relationship from the specified relationship node.
     *
     * @param entity  the parent entity
     * @param node    the relationship node
     * @param service the archetype service
     * @return the default target, or the the first target if there is no default, or {@code null} if none is found
     * @throws ArchetypeServiceException for any archetype service error
     * @throws IMObjectBeanException     if the node doesn't exist or an element is of the wrong type
     */
    public static Entity getDefaultTarget(org.openvpms.component.model.entity.Entity entity, String node,
                                          IArchetypeService service) {
        return getDefaultTarget(entity, node, true, service);
    }

    /**
     * Returns the active target from the default entity relationship from the specified relationship node.
     *
     * @param entity        the parent entity
     * @param node          the relationship node
     * @param useNonDefault if {@code true} use a non-default relationship if no default can be found
     * @param service       the archetype service
     * @return the default target, or the the first target if there is no default and {@code useNonDefault} is
     * {@code true}, or {@code null} if none is found
     * @throws ArchetypeServiceException for any archetype service error
     * @throws IMObjectBeanException     if the node doesn't exist or an element is of the wrong type
     */
    public static Entity getDefaultTarget(org.openvpms.component.model.entity.Entity entity, String node,
                                          boolean useNonDefault, IArchetypeService service) {
        IMObjectBean bean = service.getBean(entity);
        return getDefaultTarget(bean.getValues(node, Relationship.class), useNonDefault, service);
    }

    /**
     * Returns the active target from the default entity relationship from the supplied relationship list.
     *
     * @param relationships a list of relationship objects
     * @param service       the archetype service
     * @return the default target, or the the first target if there is no
     * default, or {@code null} if none is found
     * @throws ArchetypeServiceException for any archetype service error
     */
    public static Entity getDefaultTarget(List<Relationship> relationships, IArchetypeService service) {
        return getDefaultTarget(relationships, true, service);
    }

    /**
     * Returns the active target from the default entity relationship from the supplied relationship list.
     *
     * @param relationships a list of relationship objects
     * @param useNonDefault if {@code true} use a non-default relationship if no default can be found
     * @param service       the archetype service
     * @return the default target, or the the first target if there is no default and {@code useNonDefault}
     * is {@code true}, or {@code null} if none is found
     * @throws ArchetypeServiceException for any archetype service error
     */
    public static Entity getDefaultTarget(List<Relationship> relationships, boolean useNonDefault,
                                          IArchetypeService service) {
        Entity result = null;
        for (Relationship relationship : relationships) {
            if (relationship.isActive()) {
                if (result == null && useNonDefault) {
                    result = getEntity(relationship.getTarget(), service);
                } else {
                    IMObjectBean bean = service.getBean(relationship);
                    if (bean.hasNode("default") && bean.getBoolean("default")) {
                        Entity entity = getEntity(relationship.getTarget(), service);
                        if (entity != null) {
                            result = entity;
                            break;
                        }
                    }
                }
            }
        }
        return result;
    }

    /**
     * Returns the active target reference from the default entity relationship
     * from the supplied relationship list.
     * <p/>
     * Note that this can return a reference to an inactive object.
     *
     * @param relationships a list of relationship objects
     * @param service       the archetype service
     * @return the default target, or the the first target if there is no default, or {@code null} if none is found
     * @throws ArchetypeServiceException for any archetype service error
     */
    public static IMObjectReference getDefaultTargetRef(List<EntityRelationship> relationships, boolean useNonDefault,
                                                        IArchetypeService service) {
        IMObjectReference result = null;
        for (EntityRelationship relationship : relationships) {
            if (relationship.isActive()) {
                if (result == null && useNonDefault) {
                    result = relationship.getTarget();
                } else {
                    IMObjectBean bean = service.getBean(relationship);
                    if (bean.hasNode("default") && bean.getBoolean("default")) {
                        result = relationship.getTarget();
                        if (result != null) {
                            break;
                        }
                    }
                }
            }
        }
        return result;
    }

    /**
     * Makes a relationship the default, resetting the default status of any
     * other relationship associated with the specified node.
     *
     * @param entity       the entity
     * @param node         the relationship node
     * @param relationship the relationship to make the default
     * @param service      the archetype service
     * @throws ArchetypeServiceException for any archetype service error
     */
    public static void setDefault(org.openvpms.component.model.entity.Entity entity, String node,
                                  Relationship relationship, IArchetypeService service) {
        IMObjectBean bean = service.getBean(entity);
        List<Relationship> relationships = bean.getValues(node, Relationship.class);
        for (Relationship r : relationships) {
            if (r != relationship) {
                IMObjectBean relBean = service.getBean(r);
                relBean.setValue("default", false);
            }
        }
        IMObjectBean relBean = service.getBean(relationship);
        relBean.setValue("default", true);
    }

    /**
     * Returns the entity associate with a reference.
     *
     * @param ref     the reference. May be {@code null}
     * @param service the archetype service.
     * @return the corresponding entity, or {@code null}
     * @throws ArchetypeServiceException for any archetype service error
     */
    private static Entity getEntity(Reference ref, IArchetypeService service) {
        Entity result = null;
        if (ref != null) {
            result= (Entity) service.get(ref, true);
        }
        return result;
    }

}
