/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2021 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.etl.load;

import org.hibernate.SessionFactory;
import org.junit.Before;
import org.junit.Test;
import org.openvpms.archetype.rules.workflow.ScheduleTestHelper;
import org.openvpms.archetype.test.TestHelper;
import org.openvpms.archetype.test.builder.lookup.TestLookupFactory;
import org.openvpms.component.business.domain.im.act.ActRelationship;
import org.openvpms.component.business.domain.im.common.IMObject;
import org.openvpms.component.business.domain.im.common.Participation;
import org.openvpms.component.business.domain.im.party.Party;
import org.openvpms.component.business.domain.im.product.Product;
import org.openvpms.component.business.service.archetype.IArchetypeService;
import org.openvpms.component.business.service.archetype.ValidationException;
import org.openvpms.component.business.service.archetype.helper.IMObjectBean;
import org.openvpms.component.business.service.lookup.ILookupService;
import org.openvpms.component.model.lookup.Lookup;
import org.openvpms.component.model.lookup.LookupRelationship;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.AbstractJUnit4SpringContextTests;

import java.math.BigDecimal;
import java.util.List;
import java.util.Set;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;


/**
 * Tests the {@link Loader}.
 *
 * @author Tim Anderson
 */
@ContextConfiguration("/applicationContext.xml")
public class LoaderTestCase extends AbstractJUnit4SpringContextTests {

    /**
     * The lookup service.
     */
    @Autowired
    private ILookupService lookups;

    /**
     * The archetype service.
     */
    @Autowired
    private IArchetypeService service;

    /**
     * The Hibernate session factory.
     */
    @Autowired
    private SessionFactory sessionFactory;

    /**
     * The DAO.
     */
    private ETLLogDAO dao;

    /**
     * The lookup factory.
     */
    private TestLookupFactory lookupFactory;

    /**
     * Sets up the test case.
     */
    @Before
    public void setUp() {
        dao = new ETLLogDAOImpl(sessionFactory);
        lookupFactory = new TestLookupFactory(service);
    }

    /**
     * Tests a single object.
     *
     * @throws Exception for any error
     */
    @Test
    public void testObject() throws Exception {
        String loaderName = "CUSTLOAD";
        dao.remove(loaderName, "ID1");

        Mappings mappings = new Mappings();
        mappings.setIdColumn("LEGACY_ID");

        Mapping firstNameMap = createMapping("FIRST_NAME", "<party.customerperson>firstName");
        Mapping lastNameMap = createMapping("LAST_NAME", "<party.customerperson>lastName");
        mappings.addMapping(firstNameMap);
        mappings.addMapping(lastNameMap);

        Loader loader = createLoader(loaderName, mappings);
        String legacyId = "ID1";
        ETLRow row = createCustomerRow(legacyId, "Foo", "Bar");

        List<IMObject> objects = loader.load(row);
        loader.close();
        assertEquals(1, objects.size());
        IMObject object = objects.get(0);
        IMObjectBean bean = new IMObjectBean(object, service);
        assertEquals("Foo", bean.getString("firstName"));
        assertEquals("Bar", bean.getString("lastName"));

        List<ETLLog> logs = dao.get(loaderName, legacyId, null);
        assertEquals(1, logs.size());
        checkLog(logs, loaderName, legacyId, "party.customerperson", -1);
    }

    /**
     * Tests a collection node.
     */
    @Test
    public void testCollection() {
        String loaderName = "CUSTLOAD";
        String legacyId = "ID1";
        dao.remove(loaderName, legacyId);
        Mappings mappings = new Mappings();
        mappings.setIdColumn("LEGACY_ID");

        Mapping addressMap = createMapping("ADDRESS", "<party.customerperson>contacts[0]<contact.location>address");
        Mapping suburbMap = createMapping("SUBURB", "<party.customerperson>contacts[0]<contact.location>suburb");
        Mapping phoneMap = createMapping("PHONE",
                                         "<party.customerperson>contacts[1]<contact.phoneNumber>telephoneNumber");
        mappings.addMapping(addressMap);
        mappings.addMapping(suburbMap);
        mappings.addMapping(phoneMap);

        Loader loader = createLoader(loaderName, mappings);
        ETLRow row = new ETLRow(legacyId);
        row.add("ADDRESS", "49 Foo St Bar");
        row.add("SUBURB", "Coburg");
        row.add("PHONE", "123456789");

        List<IMObject> objects = loader.load(row);
        loader.close();
        assertEquals(3, objects.size());
        IMObjectBean customer = new IMObjectBean(objects.get(0));
        IMObjectBean location = new IMObjectBean(objects.get(1));
        IMObjectBean phone = new IMObjectBean(objects.get(2));
        assertTrue(customer.isA("party.customerperson"));
        assertTrue(location.isA("contact.location"));
        assertTrue(phone.isA("contact.phoneNumber"));

        List<ETLLog> logs = dao.get(loaderName, legacyId, null);
        assertEquals(1, logs.size());
        checkLog(logs, loaderName, legacyId, "party.customerperson", -1);
    }

    /**
     * Tests a collection hierarchy.
     */
    @Test
    public void testCollectionHierarchy() {
        String loaderName = "CUSTLOAD";
        dao.remove(loaderName, "ID1");

        Mappings mappings = new Mappings();
        mappings.setIdColumn("LEGACY_ID");

        lookupFactory.newLookup("lookup.contactPurpose").code("MAILING").build();
        // make sure the lookup exists

        Mapping firstNameMap = createMapping("FIRST_NAME", "<party.customerperson>firstName");
        Mapping lastNameMap = createMapping("LAST_NAME", "<party.customerperson>lastName");
        Mapping suburbMap = createMapping("ADDRESS", "<party.customerperson>contacts[0]<contact.location>purposes[1]");
        String ref = "<lookup.contactPurpose>code=MAILING";
        suburbMap.setValue(ref);
        mappings.addMapping(firstNameMap);
        mappings.addMapping(lastNameMap);
        mappings.addMapping(suburbMap);

        Loader loader = createLoader(loaderName, mappings);
        ETLRow row = createCustomerRow("ID1", "Foo", "Bar");
        row.add("ADDRESS", "49 Foo St Bar");
        List<IMObject> objects = loader.load(row);
        loader.close();

        assertEquals(2, objects.size());

        IMObjectBean customer = new IMObjectBean(objects.get(0));
        IMObjectBean location = new IMObjectBean(objects.get(1));
        assertTrue(customer.isA("party.customerperson"));
        assertTrue(location.isA("contact.location"));
    }

    /**
     * Verifies that the string <em>$value</em> is expanded with the input value.
     */
    @Test
    public void testValueExpansion() {
        Mappings mappings = new Mappings();
        mappings.setIdColumn("INVOICEID");

        Mapping mapping = createMapping(
                "INVOICEID",
                "<act.customerAccountChargesInvoice>items[0]<actRelationship.customerAccountInvoiceItem>source");
        mapping.setValue("<act.customerAccountChargesInvoice>$value");
        mappings.addMapping(mapping);

        Loader loader = createLoader("ACTLOAD", mappings);
        ETLRow row = new ETLRow("INVOICE1");
        row.add("INVOICEID", "INVOICE1");
        List<IMObject> objects = loader.load(row);
        loader.close();

        assertEquals(2, objects.size());
        IMObjectBean act = new IMObjectBean(objects.get(0));
        IMObjectBean rel = new IMObjectBean(objects.get(1));
        assertTrue(act.isA("act.customerAccountChargesInvoice"));
        assertTrue(rel.isA("actRelationship.customerAccountInvoiceItem"));
        List<ActRelationship> items
                = act.getValues("items", ActRelationship.class);
        assertEquals(items.size(), 1);
        assertEquals(rel.getObject(), items.get(0));
        assertEquals(act.getObject().getObjectReference(), rel.getValue("source"));
    }

    /**
     * Verifies that an object isn't created if 'excludeNull' is specified and the input field is null.
     */
    @Test
    public void testExcludeNull() {
        Mappings mappings = new Mappings();
        mappings.setIdColumn("CUSTID");

        Mapping mapping = createMapping("TELEPHONE",
                                        "<party.customerperson>contacts[0]<contact.phoneNumber>telephoneNumber");
        mapping.setExcludeNull(true);
        mappings.addMapping(mapping);

        Loader loader = createLoader("CUSTLOAD", mappings);
        ETLRow row = new ETLRow("CUSTID");
        row.add("TELEPHONE", null);
        List<IMObject> objects = loader.load(row);
        loader.close();

        assertEquals(0, objects.size());
    }

    /**
     * Verifies that wildcards can be specified for archetype/legacyId
     * references.
     */
    @Test
    public void testReferenceWildcards() {
        // remove any existing logs
        String loaderName = "INVLOAD";
        dao.remove("INVLOAD", "INV1");
        dao.remove("INVLOAD", "PROD1");

        // create a product
        Product product = service.create("product.medication", Product.class);
        assertNotNull(product);
        product.setName("XLoaderTestCaseProduct" + System.currentTimeMillis());
        service.save(product);

        // add a log for the product, so it can be referenced
        addLog(loaderName, "PROD1", product);

        // create a product participation mapping
        Mappings mappings = new Mappings();
        mappings.setIdColumn("INVOICEID");
        Mapping productMap = createMapping("PRODUCTID",
                                           "<act.customerAccountInvoiceItem>product[0]<participation.product>entity",
                                           "<product.*>$value");
        mappings.addMapping(productMap);

        // load a (partial) invoice item
        Loader loader = createLoader(loaderName, mappings);
        ETLRow row = new ETLRow("INVOICEID");
        row.add("INVOICEID", "INV1");
        row.add("PRODUCTID", "PROD1");
        List<IMObject> objects = loader.load(row);
        loader.close();

        // verify the invoice and participation objects were created and
        // that they reference each other
        assertEquals(2, objects.size());
        IMObjectBean act = new IMObjectBean(objects.get(0));
        IMObjectBean participation = new IMObjectBean(objects.get(1));
        assertTrue(act.isA("act.customerAccountInvoiceItem"));
        assertTrue(participation.isA("participation.product"));

        List<Participation> participations = act.getValues("product", Participation.class);
        assertEquals(1, participations.size());
        assertEquals(participations.get(0), participation.getObject());
        assertEquals(product.getObjectReference(), participation.getValue("entity"));
    }

    /**
     * Verifies that lookups and lookup relationships are saved after each
     * row loads.
     */
    @Test
    public void testLookups() {
        String loaderName = "PATIENTLOAD";
        dao.remove(loaderName, "patient1");
        dao.remove(loaderName, "patient2");

        // create a patient mapping
        Mappings mappings = new Mappings();
        mappings.setIdColumn("PATIENTID");
        Mapping nameMap = createMapping("NAME", "<party.patientpet>name");
        Mapping speciesMap = createMapping("SPECIES", "<party.patientpet>species");
        Mapping breedMap = createMapping("BREED", "<party.patientpet>breed");
        Mapping sexMap = createMapping("SEX", "<party.patientpet>sex");
        mappings.addMapping(nameMap);
        mappings.addMapping(speciesMap);
        mappings.addMapping(breedMap);
        mappings.addMapping(sexMap);

        String species1 = "S1" + System.currentTimeMillis();
        String species2 = "S2" + System.currentTimeMillis();
        String breed1 = "B1" + System.currentTimeMillis();
        String breed2 = "B2" + System.currentTimeMillis();

        // load (partial) patients and verify the lookups have been created
        Loader loader = createLoader(loaderName, mappings);
        ETLRow row1 = createPatientRow("patient1", "Fido", species1, breed1);
        List<IMObject> objects1 = loader.load(row1);
        assertEquals(1, objects1.size());
        checkSpeciesBreedLookups(species1, breed1);

        ETLRow row2 = createPatientRow("patient2", "Spot", species2, breed2);
        List<IMObject> objects2 = loader.load(row2);
        assertEquals(1, objects2.size());
        checkSpeciesBreedLookups(species2, breed2);
        loader.close();

        // verify the patients have saved, and contains the expected details
        IMObject patient1 = service.get(objects1.get(0).getObjectReference());
        IMObject patient2 = service.get(objects2.get(0).getObjectReference());
        assertNotNull(patient1);
        assertNotNull(patient2);

        IMObjectBean bean1 = new IMObjectBean(patient1);
        IMObjectBean bean2 = new IMObjectBean(patient2);
        assertEquals("Fido", bean1.getString("name"));
        assertEquals("Spot", bean2.getString("name"));
        assertEquals(species1, bean1.getString("species"));
        assertEquals(species2, bean2.getString("species"));
        assertEquals(breed1, bean1.getString("breed"));
        assertEquals(breed2, bean2.getString("breed"));
        assertEquals("Male " + breed1, bean1.getString("description"));
        assertEquals("Male " + breed2, bean2.getString("description"));
    }

    /**
     * Verifies that lookup.local codes are passed through and not uppercased.
     */
    @Test
    public void testLocalLookups() {
        Mappings mappings = new Mappings();
        mappings.setIdColumn("TEMPLATEID");
        Mapping archetypeMap = createMapping("ARCHETYPE", "<entity.documentTemplate>archetype");
        mappings.addMapping(archetypeMap);

        Loader loader = createLoader("TEMPLATELOAD", mappings);
        ETLRow row1 = createTemplateRow("document1", "act.patientDocumentForm");
        List<IMObject> objects = loader.load(row1);
        assertEquals(1, objects.size());
        assertEquals("act.patientDocumentForm", objects.get(0).getDetails().get("archetype"));

        ETLRow row2 = createTemplateRow("document2", "REPORT");
        objects = loader.load(row2);
        assertEquals(1, objects.size());
        assertEquals("REPORT", objects.get(0).getDetails().get("archetype"));
        loader.close();
    }

    /**
     * Tests mapping references.
     * <p>
     * This adds two prices to an existing product.
     */
    @Test
    public void testMappingReference() {
        String loaderName = "PRICELOAD";
        dao.remove(loaderName, "PROD1");
        dao.remove(loaderName, "ID1");
        dao.remove(loaderName, "ID2");

        // create a product
        Product product = service.create("product.medication", Product.class);
        assertNotNull(product);
        product.setName("XLoaderTestCaseProduct" + System.currentTimeMillis());
        service.save(product);

        // add a log for the product, so it can be referenced
        addLog(loaderName, "PROD1", product);

        Mappings mappings = new Mappings();
        mappings.setIdColumn("LEGACY_ID");

        Mapping priceMap = createMapping("PRICE", "$PRODID<product.medication>prices[1]<productPrice.unitPrice>price");
        mappings.addMapping(priceMap);

        Loader loader = createLoader(loaderName, mappings);

        // create two rows for different prices that reference the same product
        String priceId1 = "ID1";
        String priceId2 = "ID2";
        BigDecimal price1 = new BigDecimal("1.5");
        BigDecimal price2 = new BigDecimal("0.95");
        ETLRow row1 = createPriceRow(priceId1, "PROD1", price1);
        ETLRow row2 = createPriceRow(priceId2, "PROD1", price2);

        List<IMObject> objects1 = loader.load(row1);
        List<IMObject> objects2 = loader.load(row2);
        loader.close();
        assertEquals(2, objects1.size());
        assertEquals(product, objects1.get(0));

        assertEquals(2, objects2.size());
        assertEquals(product, objects2.get(0));

        IMObjectBean price1Bean = new IMObjectBean(objects1.get(1), service);
        assertTrue(price1Bean.isA("productPrice.unitPrice"));
        assertEquals(0, price1.compareTo(price1Bean.getBigDecimal("price")));

        IMObjectBean price2Bean = new IMObjectBean(objects2.get(1), service);
        assertTrue(price2Bean.isA("productPrice.unitPrice"));
        assertEquals(0, price2.compareTo(price2Bean.getBigDecimal("price")));

        List<ETLLog> logs = dao.get(loaderName, priceId1, null);
        assertEquals(1, logs.size());
        checkLog(logs, loaderName, priceId1, "product.medication", -1);

        logs = dao.get(loaderName, priceId2, null);
        assertEquals(1, logs.size());
        checkLog(logs, loaderName, priceId2, "product.medication", -1);
    }

    /**
     * Verifies that when mapping to a lookup node which accepts a wildcard, the mapping value can be used to specify
     * which lookup archetype to create.
     * <p/>
     * For this test, the {@code act.customerAppointment} reason node is used. This accepts {@code lookup.visitReason*}.
     * The mapping {@code <lookup.visitReason>$value} indicates to create a {@code lookup.visitReason} if it
     * doesn't exist, with $value representing the lookup code.
     */
    @Test
    public void testMappingWildcardLookup() {
        String loaderName = "APPTLOAD";
        dao.remove(loaderName, "CUST1");
        dao.remove(loaderName, "SCHEDULE1");
        dao.remove(loaderName, "APPT1");

        Lookup checkup = lookupFactory.newLookup("lookup.visitReason").uniqueCode("CHECKUP").build(false);
        // ensures the lookup doesn't already exist
        assertNull(lookups.getLookup("lookup.visitReason", checkup.getCode()));

        Party customer = TestHelper.createCustomer();
        Party location = TestHelper.createLocation();
        Party schedule = ScheduleTestHelper.createSchedule(location);

        // add a log for the customer, so it can be referenced
        addLog(loaderName, "CUST1", customer);
        addLog(loaderName, "SCHEDULE1", schedule);

        Mappings mappings = new Mappings();
        mappings.addMapping(createMapping("CUSTOMER",
                                          "<act.customerAppointment>customer[0]<participation.customer>entity",
                                          "<party.customerperson>$value"));
        mappings.addMapping(createMapping("SCHEDULE",
                                          "<act.customerAppointment>schedule[0]<participation.schedule>entity",
                                          "<party.organisationSchedule>$value"));
        mappings.addMapping(createMapping("REASON", "<act.customerAppointment>reason", "<lookup.visitReason>$value"));
        Loader loader = createLoader(loaderName, mappings);

        ETLRow row = new ETLRow("APPT1");
        row.add("CUSTOMER", "CUST1");
        row.add("SCHEDULE", "SCHEDULE1");
        row.add("REASON", checkup.getCode());

        // perform the mapping and verify the lookup was created
        loader.load(row);
        assertNotNull(lookups.getLookup("lookup.visitReason", checkup.getCode()));
    }

    /**
     * Verifies that a formatted error message is logged and that the registered
     * error handler is invoked if an object fails to be processed.
     */
    @Test
    public void testErrorLogging() {
        String loaderName = "CUSTLOAD";
        final String legacyId = "ID1";
        dao.remove(loaderName, legacyId);
        final String expectedError = "Failed to validate Last Name of Customer: value is required";
        Mappings mappings = new Mappings();
        mappings.setIdColumn("LEGACY_ID");

        Mapping firstNameMap = createMapping("FIRST_NAME", "<party.customerperson>firstName");
        Mapping lastNameMap = createMapping("LAST_NAME", "<party.customerperson>lastName");
        mappings.addMapping(firstNameMap);
        mappings.addMapping(lastNameMap);

        Loader loader = new Loader(loaderName, mappings, dao, service,
                                   lookups, new DefaultObjectHandler(loaderName, mappings, dao, service));
        ETLRow row = createCustomerRow(legacyId, "Foo", null);

        // register an error listener
        Listener listener = new Listener(legacyId, expectedError, ValidationException.class);
        loader.setErrorListener(listener);

        loader.load(row);
        loader.close();

        // verify the error listener was invoked once
        assertEquals(1, listener.getRowErrorCount());

        // verify there is a single log, with the expected error message
        checkLogError(loaderName, legacyId, expectedError);
    }

    /**
     * Verifies that failure processing one row doesn't prevent processing of
     * other rows.
     */
    @Test
    public void testPartialFailure() {
        final String loaderName = "CUSTLOAD";
        final String expectedError = "Failed to validate Last Name of Customer: value is required";

        dao.remove(loaderName, "ID1");
        dao.remove(loaderName, "ID2");
        dao.remove(loaderName, "PATIENT1");

        // create a patient
        Party patient = service.create("party.patientpet", Party.class);
        IMObjectBean bean = new IMObjectBean(patient, service);
        bean.setValue("species", "Foo");
        bean.setValue("breed", "Bar");
        patient.setName("XLoaderTestCasePatient" + System.currentTimeMillis());
        service.save(patient);

        // add a log for the patient, so it can be referenced
        addLog(loaderName, "PATIENT1", patient);

        Mappings mappings = new Mappings();
        mappings.setBatchSize(50);
        mappings.setIdColumn("CUSTOMERID");

        Mapping firstNameMap = createMapping("FIRST_NAME", "<party.customerperson>firstName");
        Mapping lastNameMap = createMapping("LAST_NAME", "<party.customerperson>lastName");
        Mapping sourceMap = createMapping("CUSTOMERID", "<entityRelationship.patientOwner>source",
                                          "<party.customerperson>$value");
        Mapping targetMap = createMapping("PATIENTID", "<entityRelationship.patientOwner>target",
                                          "<party.patientpet>$value");
        mappings.addMapping(firstNameMap);
        mappings.addMapping(lastNameMap);
        mappings.addMapping(sourceMap);
        mappings.addMapping(targetMap);

        Loader loader = new Loader(loaderName, mappings, dao, service,
                                   lookups, new DefaultObjectHandler(loaderName, mappings, dao, service));
        ETLRow row1 = createCustomerRow("ID1", "Foo", "Bar");
        row1.add("PATIENTID", "PATIENT1");
        ETLRow row2 = createCustomerRow("ID2", "Pippi", null);
        row2.add("PATIENTID", "PATIENT1");
        ETLRow row3 = createCustomerRow("ID3", "Rin", "Bar");
        row3.add("PATIENTID", "PATIENT1");

        // register an error listener that checks for failure of row ID2
        Listener listener = new Listener("ID2", expectedError, ValidationException.class);
        loader.setErrorListener(listener);

        // load the rows
        loader.load(row1);
        loader.load(row2);
        loader.load(row3);
        loader.close();

        // verify the error listener was invoked once
        assertEquals(1, listener.getRowErrorCount());

        // verify there is a 2 logs present for each row, with expected
        // values for error message
        checkLogError(loaderName, "ID1", null, 2);
        checkLogError(loaderName, "ID2", expectedError, 2);
        checkLogError(loaderName, "ID3", null, 2);
    }

    /**
     * Adds a log for an object so it can be referenced.
     *
     * @param loaderName the loader name
     * @param legacyId   the legacy row identifier
     * @param object     the object
     */
    private void addLog(String loaderName, String legacyId, IMObject object) {
        ETLLog log = new ETLLog(loaderName, legacyId, object.getArchetype());
        log.setReference(object.getObjectReference());
        dao.save(log);
    }

    /**
     * Verifies a log contains the expected data.
     *
     * @param logs       the logs
     * @param loaderName the expected loader name
     * @param legacyId   the expected legacy row identifier
     * @param archetype  the expected archetype
     * @param index      the expected index
     */
    private void checkLog(List<ETLLog> logs, String loaderName, String legacyId, String archetype, int index) {
        boolean found = false;
        for (ETLLog log : logs) {
            if (log.getLoader().equals(loaderName)
                && log.getRowId().equals(legacyId)
                && log.getArchetype().equals(archetype)
                && log.getIndex() == index) {
                found = true;
                break;
            }
        }
        assertTrue(found);
    }

    /**
     * Verifies that a single log exists for a loader and legacyId,
     * with the specified error message (or null if it isn't in error).
     *
     * @param loaderName the loader name
     * @param legacyId   the legacy row identifier
     * @param message    the expected error message. May be {@code null}
     */
    private void checkLogError(String loaderName, String legacyId, String message) {
        checkLogError(loaderName, legacyId, message, 1);
    }

    /**
     * Verifies that a the expected no. of logs exists for a loader and
     * legacyId, with the specified error message (or null if it isn't in
     * error).
     *
     * @param loaderName the loader name
     * @param legacyId   the legacy row identifier
     * @param message    the expected error message. May be {@code null}
     * @param count      the expected no. of logs
     */
    private void checkLogError(String loaderName, String legacyId, String message, int count) {
        // verify there is a single log, with the expected error message
        List<ETLLog> logs = dao.get(loaderName, legacyId, null);
        assertEquals(count, logs.size());
        for (ETLLog log : logs) {
            assertEquals(message, log.getErrors());
            if (message != null) {
                assertEquals(-1, log.getId());
            } else {
                assertTrue(log.getId() != -1);
            }
        }
    }

    /**
     * Verifies that species and breed lookups have been created, with a
     * relationship between each
     *
     * @param speciesCode the species code
     * @param breedCode   the breed code
     */
    private void checkSpeciesBreedLookups(String speciesCode, String breedCode) {
        Lookup species = lookups.getLookup("lookup.species", speciesCode);
        assertNotNull(species);
        Lookup breed = lookups.getLookup("lookup.breed", breedCode);
        assertNotNull(breed);

        // make sure the relationship exists
        Set<LookupRelationship> speciesRelationships = species.getLookupRelationships();
        assertEquals(1, speciesRelationships.size());
        LookupRelationship s = speciesRelationships.iterator().next();
        assertEquals(species.getObjectReference(), s.getSource());
        assertEquals(breed.getObjectReference(), s.getTarget());

        Set<LookupRelationship> breedRelationships = breed.getLookupRelationships();
        assertEquals(1, breedRelationships.size());
        LookupRelationship b = breedRelationships.iterator().next();
        assertEquals(species.getObjectReference(), b.getSource());
        assertEquals(breed.getObjectReference(), b.getTarget());
    }

    /**
     * Helper to create a row containing a customer first name and last name.
     *
     * @param legacyId  the customer legacy id
     * @param firstName the first name
     * @param lastName  the last name
     * @return a new row
     */
    private ETLRow createCustomerRow(String legacyId, String firstName, String lastName) {
        ETLRow row = new ETLRow(legacyId);
        row.add("CUSTOMERID", legacyId);
        row.add("FIRST_NAME", firstName);
        row.add("LAST_NAME", lastName);
        return row;
    }

    /**
     * Helper to create a row containing a patient id, species and breed.
     *
     * @param rowId   the patient id
     * @param name    the patient name
     * @param species the species
     * @param breed   the breed
     * @return a new row
     */
    private ETLRow createPatientRow(String rowId, String name, String species, String breed) {
        ETLRow row = new ETLRow(rowId);
        row.add("PATIENTID", rowId);
        row.add("NAME", name);
        row.add("SPECIES", species);
        row.add("BREED", breed);
        row.add("SEX", "MALE");
        return row;
    }

    /**
     * Helper to create a row containing a price id, product id and price.
     *
     * @param legacyId  the price legacy id
     * @param productId the product legacy id
     * @param price     the price
     * @return a new row
     */
    private ETLRow createPriceRow(String legacyId, String productId, BigDecimal price) {
        ETLRow row = new ETLRow(legacyId);
        row.add("PRICE", price);
        row.add("PRODID", productId);
        return row;
    }

    /**
     * Helper to create a row containing a document template id and archetype.
     *
     * @param rowId     the document template id
     * @param archetype the archetype
     * @return a new row
     */
    private ETLRow createTemplateRow(String rowId, String archetype) {
        ETLRow row = new ETLRow(rowId);
        row.add("TEMPLATEID", rowId);
        row.add("ARCHETYPE", archetype);
        return row;
    }

    /**
     * Helper to create a new mapping.
     *
     * @param source the source to map
     * @param target the target to map to
     * @return a new mapping
     */
    private Mapping createMapping(String source, String target) {
        return createMapping(source, target, null);
    }

    /**
     * Helper to create a new mapping.
     *
     * @param source the source to map
     * @param target the target to map to
     * @param value  the value. May be {@code null}
     * @return a new mapping
     */
    private Mapping createMapping(String source, String target, String value) {
        Mapping mapping = new Mapping();
        mapping.setSource(source);
        mapping.setTarget(target);
        mapping.setValue(value);
        return mapping;
    }

    private Loader createLoader(String loaderName, Mappings mappings) {
        return new Loader(loaderName, mappings, dao, service,
                          lookups, new DefaultObjectHandler(loaderName, mappings, dao, service));
    }

    private static class Listener implements ErrorListener {

        private final String expectedRowId;

        private final String expectedMessage;

        private final Class<?> expectedException;

        private int count;

        public Listener(String rowId, String message, Class<?> exception) {
            this.expectedRowId = rowId;
            this.expectedMessage = message;
            this.expectedException = exception;
        }

        /**
         * Returns the no. of times an error occurred for a row.
         *
         * @return the error count
         */
        public int getRowErrorCount() {
            return count;
        }

        /**
         * Invoked when an error occurs for a particular row.
         *
         * @param rowId     the identifier of the row that triggered the error
         * @param message   the error message
         * @param exception the exception
         */
        public void error(String rowId, String message, Throwable exception) {
            ++count;
            assertEquals(expectedRowId, rowId);
            assertEquals(expectedMessage, message);
            assertEquals(expectedException, exception.getClass());
        }

        /**
         * Invoked when an error occurs.
         *
         * @param message   the error message
         * @param exception the exception
         */
        public void error(String message, Throwable exception) {
            assertEquals(expectedMessage, message);
            assertEquals(expectedException, exception.getClass());
        }
    }
}
