/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.laboratory.internal.report;

import org.openvpms.archetype.rules.util.DateRules;
import org.openvpms.component.business.domain.im.act.BeanActDecorator;
import org.openvpms.component.model.act.Act;
import org.openvpms.component.model.bean.IMObjectBean;
import org.openvpms.component.model.bean.Policies;
import org.openvpms.component.model.bean.Policy;
import org.openvpms.component.model.entity.Entity;
import org.openvpms.component.model.object.SequencedRelationship;
import org.openvpms.domain.internal.factory.DomainService;
import org.openvpms.domain.laboratory.Test;
import org.openvpms.laboratory.report.Result;
import org.openvpms.laboratory.report.Results;

import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

/**
 * Default implementation of {@link Results}.
 *
 * @author Tim Anderson
 */
public class ResultsImpl implements Results {

    /**
     * The domain object service.
     */
    private final DomainService service;

    /**
     * The underlying act.
     */
    private BeanActDecorator act;

    /**
     * The results.
     */
    private List<Result> results;

    /**
     * Constructs a {@link ResultsImpl}.
     *
     * @param act     the act to delegate to
     * @param service the domain object service
     */
    public ResultsImpl(Act act, DomainService service) {
        this.act = new BeanActDecorator(act, service);
        this.service = service;
    }

    /**
     * Constructs a {@link ResultsImpl}.
     *
     * @param bean    the bean wrapping the act
     * @param service the domain object service
     */
    public ResultsImpl(IMObjectBean bean, DomainService service) {
        this.act = new BeanActDecorator(bean);
        this.service = service;
    }

    /**
     * Returns the results identifier.
     *
     * @return the results identifier
     */
    @Override
    public String getResultsId() {
        return act.getBean().getString("resultsId");
    }

    /**
     * Returns the results status.
     *
     * @return the results status
     */
    @Override
    public Result.Status getStatus() {
        return ResultImpl.getStatus(act.getStatus());
    }

    /**
     * Returns the date/time when the results where produced.
     *
     * @return the date/time
     */
    @Override
    public OffsetDateTime getDate() {
        return DateRules.toOffsetDateTime(act.getActivityStartTime());
    }

    /**
     * Returns the test.
     *
     * @return the test. May be {@code null}
     */
    @Override
    public Test getTest() {
        Entity test = act.getBean().getTarget("test", Entity.class);
        return test != null ? service.create(test, Test.class) : null;
    }

    /**
     * Returns the result category code.
     * <p/>
     * This can be used classify results that aren't associated with a test.
     *
     * @return the result category code. May be {@code null}
     */
    @Override
    public String getCategoryCode() {
        return act.getBean().getString("categoryCode");
    }

    /**
     * Returns the result category name.
     * <p/>
     * This can be used classify results that aren't associated with a test.
     *
     * @return the result category name. May be {@code null}
     */
    @Override
    public String getCategoryName() {
        return act.getBean().getString("categoryName");
    }

    /**
     * Returns a short narrative of the results.
     *
     * @return the notes. May be {@code null}
     */
    @Override
    public String getNotes() {
        return act.getBean().getString("notes");
    }

    /**
     * Returns the results.
     *
     * @return the results, in the order they were added
     */
    @Override
    public List<Result> getResults() {
        if (results == null) {
            List<Result> list = new ArrayList<>();
            Comparator<SequencedRelationship> comparator = Comparator.comparingInt(SequencedRelationship::getSequence);
            Policy<SequencedRelationship> policy = Policies.all(SequencedRelationship.class, comparator);
            for (Act act : act.getBean().getTargets("items", Act.class, policy)) {
                list.add(service.create(act, Result.class));
            }
            results = list;
        }
        return results;
    }
}
