/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.sms.message;

import java.time.OffsetDateTime;

/**
 * Builder for {@link Reply} instances.
 *
 * @author Tim Anderson
 */
public interface ReplyBuilder {

    /**
     * Sets the reply identifier, issued by the SMS provider.
     * <p>
     * A reply can have a single identifier issued by a provider. To avoid duplicates, each SMS provider must
     * use a unique archetype.
     *
     * @param archetype the identifier archetype. Must have an <em>actIdentity.sms</em> prefix.
     * @param id        the reply identifier
     * @return this
     */
    ReplyBuilder providerId(String archetype, String id);

    /**
     * Sets the timestamp when the reply was received.
     *
     * @param timestamp the received timestamp
     * @return this
     */
    ReplyBuilder received(OffsetDateTime timestamp);

    /**
     * Sets the phone number.
     *
     * @param phone the phone number
     * @return this
     */
    ReplyBuilder phone(String phone);

    /**
     * Sets the message.
     *
     * @param message the message
     * @return this
     */
    ReplyBuilder message(String message);

    /**
     * Builds the reply.
     *
     * @return the reply
     */
    Reply build();

    /**
     * Builds the reply, and returns the state builder.
     * <p/>
     * This can be used when performing multiple changes to a message.
     *
     * @return the message state builder
     */
    StateBuilder add();

}
