/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.sms.internal.message;

import org.openvpms.archetype.rules.util.DateRules;
import org.openvpms.component.business.service.archetype.helper.TypeHelper;
import org.openvpms.component.model.act.Act;
import org.openvpms.component.model.act.ActIdentity;
import org.openvpms.component.model.bean.IMObjectBean;
import org.openvpms.component.service.archetype.ArchetypeService;
import org.openvpms.sms.internal.SMSArchetypes;
import org.openvpms.sms.message.Reply;
import org.openvpms.sms.message.ReplyBuilder;
import org.openvpms.sms.message.StateBuilder;

import java.time.OffsetDateTime;

/**
 * Default implementation of {@link ReplyBuilder}.
 *
 * @author Tim Anderson
 */
public class ReplyBuilderImpl implements ReplyBuilder {

    /**
     * The parent builder.
     */
    private final StateBuilderImpl parent;

    /**
     * The archetype service.
     */
    private final ArchetypeService service;

    /**
     * The identifier archetype. Must have an <em>actIdentity.sms</em> prefix.
     */
    private String archetype;

    /**
     * The message identifier.
     */
    private String id;

    /**
     * The time when the reply was received.
     */
    private OffsetDateTime received;

    /**
     * The phone.
     */
    private String phone;

    /**
     * The message.
     */
    private String message;

    /**
     * Constructs a {@link ReplyBuilderImpl}.
     *
     * @param parent  the parent builder
     * @param service the archetype service
     */
    public ReplyBuilderImpl(StateBuilderImpl parent, ArchetypeService service) {
        this.parent = parent;
        this.service = service;
    }

    /**
     * Sets the reply identifier, issued by the SMS provider.
     * <p>
     * A reply can have a single identifier issued by a provider. To avoid duplicates, each SMS provider must
     * use a unique archetype.
     *
     * @param archetype the identifier archetype. Must have an <em>actIdentity.sms</em> prefix.
     * @param id        the reply identifier
     * @return this
     */
    @Override
    public ReplyBuilder providerId(String archetype, String id) {
        this.archetype = archetype;
        this.id = id;
        return this;
    }

    /**
     * Sets the timestamp when the reply was received.
     *
     * @param timestamp the received timestamp
     * @return this
     */
    @Override
    public ReplyBuilder received(OffsetDateTime timestamp) {
        received = timestamp;
        return this;
    }

    /**
     * Sets the phone number.
     *
     * @param phone the phone number
     * @return this
     */
    @Override
    public ReplyBuilder phone(String phone) {
        this.phone = phone;
        return this;
    }

    /**
     * Sets the message.
     *
     * @param message the message
     * @return this
     */
    @Override
    public ReplyBuilder message(String message) {
        this.message = message;
        return this;
    }

    /**
     * Builds the reply.
     *
     * @return the reply
     */
    @Override
    public Reply build() {
        Act act = prepare();
        IMObjectBean bean = service.getBean(act);
        build(act, bean);
        parent.build();
        return new ReplyImpl(bean);
    }

    /**
     * Adds the reply to the original message.
     * <p/>
     * This can be used when performing multiple changes to a message.
     *
     * @return the message state builder
     */
    @Override
    public StateBuilder add() {
        Act act = prepare();
        IMObjectBean bean = service.getBean(act);
        build(act, bean);
        return parent;
    }

    /**
     * Prepares for building.
     *
     * @return a new reply act
     */
    private Act prepare() {
        if (archetype != null && id != null) {
            if (!TypeHelper.isA(archetype, "actIdentity.sms*")) {
                throw new IllegalStateException("Invalid identity archetype: " + archetype);
            }
            parent.checkDuplicateReply(id);
        }
        return service.create(SMSArchetypes.REPLY, Act.class);
    }

    /**
     * Builds the reply.
     *
     * @param act  the reply act
     * @param bean the reply bean
     */
    private void build(Act act, IMObjectBean bean) {
        if (archetype != null && id != null) {
            ActIdentity identity = service.create(archetype, ActIdentity.class);
            identity.setIdentity(id);
            act.addIdentity(identity);
        }
        if (received != null) {
            bean.setValue("startTime", DateRules.toDate(received));
        }
        bean.setValue("phone", phone);
        bean.setValue("message", message);
        parent.addReply(act);
    }
}
