
/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2021 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.tool.db;


import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Switch;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.time.StopWatch;
import org.flywaydb.core.api.MigrationInfo;
import org.flywaydb.core.api.callback.BaseFlywayCallback;
import org.flywaydb.core.api.callback.FlywayCallback;
import org.flywaydb.core.internal.info.MigrationInfoDumper;
import org.flywaydb.core.internal.util.TimeFormat;
import org.openvpms.db.service.impl.DatabaseServiceImpl;
import org.openvpms.tool.toolbox.config.ConfigCreator;
import org.openvpms.tool.toolbox.config.ConfigProperties;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.Console;
import java.math.BigDecimal;
import java.nio.file.Paths;
import java.sql.Connection;
import java.sql.SQLException;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * Tool to create and migrate OpenVPMS databases.
 *
 * @author Tim Anderson
 */
public class DBTool {

    /**
     * The OpenVPMS user.
     */
    private final String user;

    /**
     * The OpenVPMS user password.
     */
    private final String password;

    /**
     * The database service.
     */
    private final DatabaseServiceImpl service;

    /**
     * The logger.
     */
    private static final Logger log = LoggerFactory.getLogger(DBTool.class);

    /**
     * Constructs a {@link DBTool}.
     *
     * @param driver   the driver class name
     * @param url      the driver url
     * @param user     the database user name
     * @param password the database password
     */
    public DBTool(String driver, String url, String user, String password) {
        this(driver, url, user, password, new FlywayLogger());
    }

    /**
     * Constructs a {@link DBTool}.
     *
     * @param driver   the driver class name
     * @param url      the driver url
     * @param user     the database user name
     * @param password the database password
     * @param listener the listener to notify of Flyway events. May be {@code null}
     */
    public DBTool(String driver, String url, String user, String password, FlywayCallback listener) {
        this.user = user;
        this.password = password;
        service = new DatabaseServiceImpl(driver, url, user, password, listener);
    }

    /**
     * Creates the database and tables, if it doesn't already exist.
     *
     * @param adminUser     the admin user
     * @param adminPassword the admin password
     * @param createTables  if {@code true}, create the tables, and base-line
     * @throws SQLException for any SQL error
     */
    public void create(String adminUser, String adminPassword, boolean createTables) throws SQLException {
        service.create(adminUser, adminPassword, user, password, createTables);
        System.out.println("Created " + service.getSchemaName());
    }

    /**
     * Repair database version meta-data.
     */
    public void repair() {
        service.repair();
    }

    /**
     * Displays the database version.
     */
    public void version() {
        String version = service.getVersion();
        if (version == null) {
            System.out.println("Database '" + service.getSchemaName() + "' has no version information");
        } else {
            System.out.println("Database '" + service.getSchemaName() + "' is at version " + version);
        }
    }

    /**
     * Displays migration info.
     */
    public void info() {
        System.out.println(MigrationInfoDumper.dumpToAsciiTable(service.getInfo().all()));
    }

    /**
     * Determines if the database needs updating.
     *
     * @return {@code true} if the database needs updating
     */
    public boolean needsUpdate() {
        return service.needsUpdate();
    }

    /**
     * Updates the database to the latest version.
     *
     * @throws SQLException for any SQL error
     */
    public void update() throws SQLException {
        String version = service.getVersion();
        String schemaName = service.getSchemaName();
        if (service.needsUpdate()) {
            StopWatch stopWatch = StopWatch.createStarted();
            service.update();
            if (version == null) {
                // no version information
                System.out.println("Database '" + schemaName + "' updated to version " + service.getVersion()
                                   + " in " + stopWatch);
            } else {
                System.out.println("Database '" + schemaName + "' updated from version " + version + " to "
                                   + service.getVersion() + " in " + stopWatch);
            }
        } else {
            System.out.println("Database '" + schemaName + "' is up to date");
        }
    }

    /**
     * Displays the database size.
     *
     * @throws SQLException for any SQL error
     */
    public void size() throws SQLException {
        long size = service.getSize();
        String value;
        if (size / (10 * FileUtils.ONE_GB) > 0) {
            value = getSize(size, FileUtils.ONE_GB, "GB");
        } else if (size / FileUtils.ONE_MB > 0) {
            value = getSize(size, FileUtils.ONE_MB, "MB");
        } else if (size / FileUtils.ONE_KB > 0) {
            value = getSize(size, FileUtils.ONE_KB, "KB");
        } else {
            value = size + " bytes";
        }
        System.out.println(service.getSchemaName() + " " + value);
    }

    /**
     * The main line
     *
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        try {
            JSAP parser = createParser();
            JSAPResult commands = parser.parse(args);
            if (!commands.success() || commands.getBoolean("help") || commands.getString("properties") == null) {
                displayUsage(commands);
            } else if (commands.getBoolean("config")) {
                String path = commands.getString("properties");
                ConfigCreator creator = new ConfigCreator();
                creator.create(Paths.get(path));
            } else {
                String path = commands.getString("properties");
                ConfigProperties config = new ConfigProperties(Paths.get(path));
                if (!config.isComplete()) {
                    System.err.printf("The %s configuration is incomplete. Run dbtool --config\n", path);
                    System.exit(1);
                }
                String driver = config.getDriver();
                String url = config.getUrl();
                String user = config.getUsername();
                String password = config.getPassword();
                String key = config.getKey();

                DBTool tool = new DBTool(driver, url, user, password, new FlywayLogger());
                String create = commands.getString("create");
                if (create != null && (user = commands.getString("user")) != null) {
                    boolean install = "install".equals(create);
                    boolean restore = "restore".equals(create);
                    if (install || restore) {
                        password = commands.getString("password");
                        if (StringUtils.isEmpty(password)) {
                            Console console = System.console();
                            if (console == null) {
                                System.err.println("This command must be executed in a console");
                                System.exit(1);
                            }
                            console.printf("Enter password: ");
                            char[] pass;
                            while ((pass = console.readPassword()) == null || pass.length == 0) {
                                console.printf("Enter password: ");
                            }
                            password = new String(pass);
                            Arrays.fill(pass, ' ');
                        }
                        tool.create(user, password, install);
                    } else {
                        displayUsage(commands);
                    }
                } else if (commands.getBoolean("info")) {
                    tool.info();
                } else if (commands.getBoolean("update")) {
                    System.setProperty("openvpms.key", key);
                    if (tool.needsUpdate()) {
                        if (!commands.getBoolean("backedup")) {
                            Console console = System.console();
                            if (console == null) {
                                System.err.println("This command must be executed in a console");
                                System.exit(1);
                            }
                            boolean done = false;
                            while (!done) {
                                console.printf("Has the database been backed up? [Y/n] ");
                                String input = console.readLine();
                                if ("Y".equals(input)) {
                                    done = true;
                                } else if ("n".equals(input)) {
                                    System.err.println("The database update can only be reverted by restoring a " +
                                                       "backup.\nBack up the database before re-running this command "
                                                       + "again");
                                    System.exit(1);
                                }
                            }
                        }
                        tool.update();
                    } else {
                        System.out.println("Database is up to date");
                    }
                } else if (commands.getBoolean("version")) {
                    tool.version();
                } else if (commands.getBoolean("repair")) {
                    tool.repair();
                } else {
                    displayUsage(commands);
                }
            }
        } catch (Throwable throwable) {
            log.error(throwable.getMessage(), throwable);
        }
    }

    /**
     * Creates a new command line parser.
     *
     * @return a new parser
     * @throws JSAPException if the parser can't be created
     */
    private static JSAP createParser() throws JSAPException {
        JSAP parser = new JSAP();
        parser.registerParameter(new Switch("config").setLongFlag("config").setDefault("false").setHelp(
                "Create or update the openvpms.properties configuration file"));
        parser.registerParameter(new FlaggedOption("create").setLongFlag("create").setDefault("false").setHelp(
                "Create the OpenVPMS database."));
        parser.registerParameter(new FlaggedOption("user").setShortFlag('u').setHelp("Admin user to create database"));
        parser.registerParameter(new FlaggedOption("password").setShortFlag('p')
                                         .setHelp("Admin password to create database"));
        parser.registerParameter(new Switch("info").setShortFlag('i')
                                         .setLongFlag("info").setDefault("false").setHelp(
                        "Displays migration info."));
        parser.registerParameter(new Switch("update").setLongFlag("update").setDefault("false").setHelp(
                "Updates the database to the latest version."));
        parser.registerParameter(new Switch("version").setShortFlag('v')
                                         .setLongFlag("version").setDefault("false").setHelp(
                        "Displays the database version."));
        parser.registerParameter(new Switch("help").setLongFlag("help").setDefault("false")
                                         .setHelp("Displays this help."));
        parser.registerParameter(new Switch("backedup").setLongFlag("database-is-backed-up").setDefault("false")
                                         .setHelp("If specified, disables backup prompting when updating the database"));
        parser.registerParameter(new FlaggedOption("properties").setLongFlag("properties")
                                         .setHelp("openvpms.properties path"));
        parser.registerParameter(new Switch("repair").setLongFlag("repair").setDefault("false").setHelp(
                "Repair version meta data."));
        return parser;
    }

    /**
     * Prints usage information and exits.
     *
     * @param result the parse result
     */
    private static void displayUsage(JSAPResult result) {
        Iterator iter = result.getErrorMessageIterator();
        while (iter.hasNext()) {
            System.err.println(iter.next());
        }
        System.err.println();
        System.err.println("Usage: dbtool [options]");
        System.err.println();
        System.err.println("  --config");
        System.err.println("    Creates or updates the openvpms.properties configuration file");
        System.err.println();
        System.err.println("  --create <install | restore> -u <user> [-p <password>]");
        System.err.println("    Creates the database. When:");
        System.err.println("    . install is specified, the database and tables will be created and version " +
                           "information added.");
        System.err.println("      Use this for new installations.");
        System.err.println("    . restore is specified, an empty database will be created.");
        System.err.println("      Use this when restoring backups to a new server");
        System.err.println();
        System.err.println("  --update");
        System.err.println("    Updates the database to the latest version");
        System.err.println();
        System.err.println("  --version");
        System.err.println("    Displays database version");
        System.err.println();
        System.err.println("  --info");
        System.err.println("    Displays database migration information");
        System.err.println();
        System.err.println("  --properties <path>");
        System.err.println("    Specifies the path to the openvpms.properties configuration file");
        System.err.println();
        System.err.println("  --help");
        System.err.println("    Displays this help");
        System.exit(1);
    }

    private static class FlywayLogger extends BaseFlywayCallback {

        private final Map<MigrationInfo, StopWatch> state = new HashMap<>();

        @Override
        public void beforeEachMigrate(Connection connection, MigrationInfo info) {
            StopWatch watch = new StopWatch();
            state.put(info, watch);
            System.out.print("Updating to " + info.getVersion() + " - " + info.getDescription() + " ... ");
            watch.start();
        }

        @Override
        public void afterEachMigrate(Connection connection, MigrationInfo info) {
            String time = "";
            StopWatch watch = state.get(info);
            if (watch != null) {
                watch.stop();
                time = TimeFormat.format(watch.getTime());
                System.out.println(time);
            }
        }
    }

    /**
     * Helper to return a formatted size, rounded.
     *
     * @param size    the size
     * @param divisor the divisor
     * @param suffix the size suffix
     * @return the formatted size
     */
    private String getSize(long size, long divisor, String suffix) {
        BigDecimal result = new BigDecimal(size).divide(BigDecimal.valueOf(divisor), 2, BigDecimal.ROUND_CEILING);
        NumberFormat format = new DecimalFormat("#,##.##");
        return format.format(result) + " " + suffix;
    }
}
