/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.tool.toolbox.config;

import org.apache.commons.lang3.StringUtils;
import org.openvpms.component.system.common.crypto.KeyGenerator;
import org.openvpms.tool.toolbox.util.PathHelper;

import java.io.Console;
import java.io.IOException;
import java.nio.file.Path;

/**
 * Creates or updates an openvpms.properties file.
 *
 * @author Tim Anderson
 */
public class ConfigCreator {

    /**
     * Creates or updates openvpms.properties.
     *
     * @param path the openvpms.properties file path. If {@code null}, uses the default.
     * @throws IOException for any I/O error
     */
    public void create(Path path) throws IOException {
        Console console = System.console();
        if (console == null) {
            throw new IOException("This command must be executed in a console");
        }
        ConfigPropertiesUpdater config = new ConfigPropertiesUpdater(path);
        config.setUrl(prompt(console, "Database URL", config.getUrl()));
        config.setUsername(prompt(console, "Database user", config.getUsername()));
        config.setPassword(prompt(console, "Database password", config.getPassword()));
        config.setReportingUrl(prompt(console, "Reporting database URL",
                                      config.getReportingUrl() != null ? config.getReportingUrl() : config.getUrl()));
        config.setReportingUsername(prompt(console, "Reporting database user",
                                           config.getReportingUsername() != null ? config.getReportingUsername()
                                                                                 : config.getUsername()));
        config.setReportingPassword(prompt(console, "Reporting database password",
                                           config.getReportingPassword() != null ? config.getReportingPassword()
                                                                                 : config.getPassword()));

        if (StringUtils.isEmpty(config.getKey())) {
            config.setKey(KeyGenerator.generate(14, 32));
        }

        if (config.isModified()) {
            config.write();
        }
        Path relative = PathHelper.getPathRelativeToCWD(path);
        try {
            config.setPermissions();
        } catch (Exception exception) {
            System.err.printf("\nThe file %s should have restrictive permissions to protect passwords\n", relative);
        }

        System.out.printf("\nTake a secure backup of the file %s\nThis needs to be kept for subsequent updates.\n\n",
                          relative);
    }

    /**
     * Prompts for user input and returns the entered value.
     *
     * @param console      the console
     * @param prompt       the prompt
     * @param defaultValue the default value. May be {@code null}
     * @return the entered value
     */
    private String prompt(Console console, String prompt, String defaultValue) {
        String result = null;
        boolean done = false;
        while (!done) {
            String line = defaultValue != null ? prompt + " [" + defaultValue + "]: " : prompt + ": ";
            console.printf(line);
            result = console.readLine();
            if (!StringUtils.isEmpty(result)) {
                done = true;
            } else if (defaultValue != null) {
                result = defaultValue;
                done = true;
            }
        }
        return result;
    }
}
