/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.component.property;

import nextapp.echo2.app.event.WindowPaneListener;
import org.openvpms.component.business.domain.im.common.IMObject;
import org.openvpms.component.business.service.archetype.IArchetypeService;
import org.openvpms.component.business.service.archetype.ValidationException;
import org.openvpms.component.exception.OpenVPMSException;
import org.openvpms.component.service.archetype.ValidationError;
import org.openvpms.web.echo.error.ErrorHandler;
import org.openvpms.web.resource.i18n.Messages;
import org.openvpms.web.system.ServiceHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;


/**
 * Validation helper.
 *
 * @author Tim Anderson
 */
public class ValidationHelper {

    /**
     * The logger.
     */
    private static final Logger log = LoggerFactory.getLogger(ValidationHelper.class);


    /**
     * Validates an object.
     *
     * @param object the object to validate
     * @return {@code true} if the object is valid; otherwise {@code false}
     */
    public static boolean isValid(IMObject object) {
        IArchetypeService service = ServiceHelper.getArchetypeService();
        return isValid(object, service);
    }

    /**
     * Validates an object.
     *
     * @param object  the object to validate
     * @param service the archetype service
     * @return {@code true} if the object is valid; otherwise {@code false}
     */
    public static boolean isValid(IMObject object, IArchetypeService service) {
        List<ValidatorError> errors = validate(object, service);
        return (errors == null);
    }

    /**
     * Validates an object.
     *
     * @param object  the object to validate
     * @param service the archetype service
     * @return a list of validation errors, or {@code null} if the object is valid
     */
    public static List<ValidatorError> validate(IMObject object, IArchetypeService service) {
        List<ValidatorError> result = null;
        try {
            service.validateObject(object);
        } catch (ValidationException exception) {
            log.debug(exception.getMessage(), exception);
            List<ValidationError> errors = exception.getErrors();
            result = new ArrayList<>();
            if (errors.isEmpty()) {
                result.add(new ValidatorError(exception.getMessage()));
            } else {
                for (ValidationError error : errors) {
                    result.add(new ValidatorError(error));
                }
            }
        } catch (OpenVPMSException exception) {
            log.debug(exception.getMessage(), exception);
            result = new ArrayList<>();
            result.add(new ValidatorError(exception.getMessage()));
        }
        return result;
    }

    /**
     * Displays the first error from a validator.
     * <p>
     * The error will be formatted.
     *
     * @param validator the validator
     */
    public static void showError(Validator validator) {
        showError(null, validator, null);
    }

    /**
     * Displays the first error from a validator.
     * <p>
     * The error will be formatted.
     *
     * @param title     the dialog title. May  be {@code null}
     * @param validator the validator
     * @param listener  the listener to notify when the dialog closes. May be {@code null}
     */
    public static void showError(String title, Validator validator, WindowPaneListener listener) {
        showError(title, validator, null, true, listener);
    }

    /**
     * Display the first error from a validator.
     *
     * @param title     the dialog title. May  be {@code null}
     * @param validator the validator
     * @param key       resource bundle key, if the error should be included in text. May be {@code null}
     * @param formatted if {@code true} format the message, otherwise display as is
     */
    public static void showError(String title, Validator validator, String key, boolean formatted) {
        showError(title, validator, key, formatted, null);
    }

    /**
     * Display the first error from a validator.
     *
     * @param title     the dialog title. May  be {@code null}
     * @param validator the validator
     * @param key       resource bundle key, if the error should be included in text. May be {@code null}
     * @param formatted if {@code true} format the message, otherwise display as is
     * @param listener  the listener to notify when the dialog closes. May be {@code null}
     */
    public static void showError(String title, Validator validator, String key, boolean formatted,
                                 WindowPaneListener listener) {
        ValidatorError error = validator.getFirstError();
        if (error != null) {
            String message = (formatted) ? error.toString() : error.getMessage();
            if (key != null) {
                message = Messages.format(key, message);
            }
            ErrorHandler.getInstance().error(title, message, null, listener);
        }
    }

}
