/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2021 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.echo.dialog;

import nextapp.echo2.app.event.WindowPaneListener;
import org.openvpms.web.echo.help.HelpContext;

/**
 * Builder for {@link MessageDialog}s.
 *
 * @author Tim Anderson
 */
public abstract class MessageDialogBuilder<T extends MessageDialog, B extends MessageDialogBuilder<T, B>> {

    /**
     * The dialog title.
     */
    private String title;

    /**
     * The preamble.
     */
    private String preamble;

    /**
     * The message.
     */
    private String message;

    /**
     * Determines if the message is an HTML fragment or plain text.
     */
    private boolean messageHTML;

    /**
     * The footer.
     */
    private String footer;

    /**
     * The buttons.
     */
    private String[] buttons;

    /**
     * The dialog style name.
     */
    private String style;

    /**
     * The dialog size. This overrides that of the style.
     */
    private String size;

    /**
     * The help context.
     */
    private HelpContext help;

    /**
     * The window pane listener.
     */
    private WindowPaneListener listener;

    /**
     * Sets the dialog title.
     *
     * @param title the title
     * @return this
     */
    public B title(String title) {
        this.title = title;
        return getThis();
    }

    /**
     * Returns the dialog title.
     *
     * @return the title. May be {@code null}
     */
    public String getTitle() {
        return title;
    }

    /**
     * Sets the preamble.
     *
     * @param preamble the preamble
     * @return this
     */
    public B preamble(String preamble) {
        this.preamble = preamble;
        return getThis();
    }

    /**
     * Returns the preamble.
     *
     * @return the preamble. May be {@code null}
     */
    public String getPreamble() {
        return preamble;
    }

    /**
     * Sets the message.
     *
     * @param message the message
     * @return this
     */
    public B message(String message) {
        return message(message, false);
    }

    /**
     * Sets the message.
     *
     * @param message the message
     * @param html    if {@code true}, the message is an HTML fragment, otherwise it is plain text
     * @return this
     */
    public B message(String message, boolean html) {
        this.message = message;
        this.messageHTML = html;
        return getThis();
    }

    /**
     * Returns the message.
     *
     * @return the message. May be {@code null}
     */
    public String getMessage() {
        return message;
    }

    /**
     * Determines if the message is an HTML fragment.
     *
     * @return {@code true} if the message is an HTML fragment, {@code false} if it is plain text
     */
    public boolean isMessageHTML() {
        return messageHTML;
    }

    /**
     * Sets the footer.
     *
     * @param footer the footer
     * @return this
     */
    public B footer(String footer) {
        this.footer = footer;
        return getThis();
    }

    /**
     * Returns the footer.
     *
     * @return the footer. May be {@code null}
     */
    public String getFooter() {
        return footer;
    }

    /**
     * Sets the display of YES/NO buttons.
     *
     * @return this
     */
    public B yesNo() {
        return buttons(ConfirmationDialog.YES_NO);
    }

    /**
     * Sets the buttons to display.
     *
     * @param buttons the button identifiers
     * @return this
     */
    public B buttons(String... buttons) {
        this.buttons = buttons;
        return getThis();
    }

    /**
     * Returns the button identifiers.
     *
     * @return the button identifiers
     */
    public String[] getButtons() {
        return buttons == null ? new String[0] : buttons;
    }

    /**
     * Sets the dialog style.
     *
     * @param style the dialog style
     */
    public void style(String style) {
        this.style = style;
    }

    /**
     * Returns the dialog style.
     *
     * @return the dialog style. May be {@code null}
     */
    public String getStyle() {
        return style;
    }

    /**
     * Sets the dialog size.
     *
     * @param size the dialog size property
     * @return this
     * @see PopupDialog#resize()
     */
    public B size(String size) {
        this.size = size;
        return getThis();
    }

    /**
     * Returns the dialog size.
     *
     * @return the dialog size. May be {@code null}
     */
    public String getSize() {
        return size;
    }

    /**
     * Sets the help context.
     *
     * @param help the help context
     * @return this
     */
    public B help(HelpContext help) {
        this.help = help;
        return getThis();
    }

    /**
     * Returns the help context.
     *
     * @return the help context. May be {@code null}
     */
    public HelpContext getHelp() {
        return help;
    }

    /**
     * Sets the dialog listener.
     *
     * @param listener the listener
     * @return this
     */
    public B listener(WindowPaneListener listener) {
        this.listener = listener;
        return getThis();
    }

    /**
     * Returns the dialog listener.
     *
     * @return the dialog listener. May be {@code null}
     */
    public WindowPaneListener getListener() {
        return listener;
    }

    /**
     * Builds the dialog.
     *
     * @return the dialog
     */
    public abstract T build();

    /**
     * Builds and displays the dialog.
     *
     * @return the dialog
     */
    public T show() {
        T dialog = build();
        dialog.show();
        return dialog;
    }

    /**
     * Builds and displays the dialog, automatically closing it after a period of time.
     *
     * @param time the time to show the window for, in seconds
     * @return the dialog
     */
    public T show(int time) {
        T dialog = build();
        dialog.show(time);
        return dialog;
    }

    /**
     * Returns this.
     *
     * @return this
     */
    protected abstract B getThis();
}
