/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2021 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.echo.util;

import nextapp.echo2.app.ApplicationInstance;
import nextapp.echo2.app.WindowPane;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.openvpms.web.echo.spring.SpringApplicationInstance;

/**
 * A {@link Runnable} that runs an underlying instance via a {@link TaskQueue}.<br/>
 * This allows the {@link Runnable} to update the user interface.
 * <p/>
 * Note that instances created prior to a {@link WindowPane} being displayed will have their execution suspended until
 * the dialog is closed, to avoid unwanted UI synchronisation events being generated<br/>
 * By default, any invocations to {@link #run()} during this time will be discarded, unless
 * {@code queueLastTaskOnStop} is set {@code true} on construction. This queues the last received call to {@link #run()}
 * and schedules it for execution when the dialog closes.
 *
 * @author Tim Anderson
 * @see TaskQueue
 * @see TaskQueues
 */
public class ApplicationInstanceRunnable implements Runnable {

    /**
     * The runnable to delegate to.
     */
    private final Runnable runnable;

    /**
     * The task queue.
     */
    private final TaskQueue taskQueue;

    /**
     * The logger.
     */
    private static final Log log = LogFactory.getLog(ApplicationInstanceRunnable.class);


    /**
     * Constructs an {@link ApplicationInstanceRunnable}.
     *
     * @param runnable the runnable
     */
    public ApplicationInstanceRunnable(Runnable runnable) {
        this(false, runnable);
    }

    /**
     * Constructs an {@link ApplicationInstanceRunnable}.
     *
     * @param queueLastTaskOnStop if {@code true}, keep the last task received after stopping and queue on restart
     * @param runnable            the runnable
     */
    public ApplicationInstanceRunnable(boolean queueLastTaskOnStop, Runnable runnable) {
        this.runnable = runnable;
        SpringApplicationInstance app = (SpringApplicationInstance) ApplicationInstance.getActive();
        if (app == null) {
            throw new IllegalStateException("No current ApplicationInstance");
        }
        taskQueue = app.getTaskQueues().newQueue(queueLastTaskOnStop);
    }

    /**
     * Schedules the runnable.
     */
    @Override
    public void run() {
        taskQueue.queue(this::execute);
    }

    /**
     * Disposes of this instance.
     */
    public void dispose() {
        taskQueue.dispose();
    }

    /**
     * Executes the runnable.
     */
    private void execute() {
        try {
            runnable.run();
        } catch (Exception exception) {
            log.warn("ApplicationInstanceRunnable callback threw exception", exception);
        }
    }

}
