/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2019 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.workspace.admin.organisation;

import org.openvpms.component.business.domain.im.common.IMObject;
import org.openvpms.mapping.service.MappingProvider;
import org.openvpms.plugin.manager.PluginManager;
import org.openvpms.plugin.service.config.ConfigurableService;
import org.openvpms.web.component.im.edit.IMObjectEditor;
import org.openvpms.web.component.im.layout.LayoutContext;
import org.openvpms.web.system.ServiceHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * An editor for <em>entity.plugin*</em> plugin configurations.
 * <p/>
 * This supports plugins that implement the {@link MappingProvider} interface; a mapping tab will
 * be added when the plugin is configured.
 *
 * @author Tim Anderson
 */
public class PluginEditor extends AbstractMappingPluginEditor {

    /**
     * The logger.
     */
    private static final Logger log = LoggerFactory.getLogger(PluginEditor.class);

    /**
     * Constructs a {@link PluginEditor}.
     *
     * @param object        the object to edit
     * @param parent        the parent object. May be {@code null}
     * @param layoutContext the layout context
     */
    public PluginEditor(IMObject object, IMObject parent, LayoutContext layoutContext) {
        super(object, parent, layoutContext);
    }

    /**
     * Creates a new instance of the editor, with the latest instance of the object to edit.
     *
     * @return a new instance of the editor
     */
    @Override
    public IMObjectEditor newInstance() {
        return new PluginEditor(reload(getObject()), getParent(), getLayoutContext());
    }

    /**
     * Determines if the plugin configuration is associated with a {@link MappingProvider}.
     *
     * @param config the plugin configuration
     * @return the mapping provider, or {@code null} if the service doesn't implement the {@link MappingProvider}
     * interface
     */
    @Override
    protected MappingProvider getMappingProvider(IMObject config) {
        MappingProvider result = null;
        try {
            PluginManager manager = ServiceHelper.getBean(PluginManager.class);
            for (ConfigurableService service : manager.getServices(ConfigurableService.class)) {
                if (service instanceof MappingProvider) {
                    if (configures(config, service)) {
                        result = (MappingProvider) service;
                        break;
                    }
                }
            }
        } catch (Throwable exception) {
            log.warn(exception.getMessage(), exception);
        }
        return result;
    }

    /**
     * Determines if a configuration is used to configure the specified service.
     *
     * @param config  the plugin configuration
     * @param service the configuration service
     * @return {@code true} if the configuration configures the service
     */
    private boolean configures(IMObject config, ConfigurableService service) {
        boolean result = false;
        try {
            org.openvpms.component.model.object.IMObject configuration = service.getConfiguration();
            result = (configuration != null && configuration.equals(config));
        } catch (Throwable exception) {
            log.warn(exception.getMessage(), exception);
        }
        return result;
    }

}
