/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.workspace.patient.investigation;

import nextapp.echo2.app.Component;
import nextapp.echo2.app.Grid;
import org.apache.commons.lang.StringUtils;
import org.openvpms.component.business.domain.im.common.IMObject;
import org.openvpms.web.component.im.layout.AbstractLayoutStrategy;
import org.openvpms.web.component.im.layout.ArchetypeNodes;
import org.openvpms.web.component.im.layout.ComponentGrid;
import org.openvpms.web.component.im.layout.LayoutContext;
import org.openvpms.web.component.im.view.ComponentState;
import org.openvpms.web.component.property.Property;
import org.openvpms.web.component.property.PropertySet;
import org.openvpms.web.echo.factory.ColumnFactory;
import org.openvpms.web.echo.style.Styles;

import java.util.List;

/**
 * Layout strategy for <em>act.patientInvestigationResultItem</em>.
 *
 * @author Tim Anderson
 */
public class ResultLayoutStrategy extends AbstractLayoutStrategy {

    /**
     * The result node name.
     */
    private static final String RESULT = "result";

    /**
     * The notes node name.
     */
    private static final String NOTES = "notes";

    /**
     * The nodes to display.
     */
    private static final ArchetypeNodes NODES = ArchetypeNodes.all().excludeIfEmpty(
            "value", "units", "qualifier", "extremeLowRange", "lowRange", "highRange", "extremeHighRange",
            "outOfRange", "referenceRange").exclude(NOTES);

    /**
     * Constructs a {@link ResultLayoutStrategy}.
     */
    public ResultLayoutStrategy() {
        setArchetypeNodes(NODES);
    }

    /**
     * Apply the layout strategy.
     * <p>
     * This renders an object in a {@code Component}, using a factory to create the child components.
     *
     * @param object     the object to apply
     * @param properties the object's properties
     * @param parent     the parent object. May be {@code null}
     * @param context    the layout context
     * @return the component containing the rendered {@code object}
     */
    @Override
    public ComponentState apply(IMObject object, PropertySet properties, IMObject parent, LayoutContext context) {
        Property result = properties.get(RESULT);
        if (getLines(result) > 1) {
            // if the result is multi-line create a text area to display it.
            addComponent(createMultiLineText(result, 2, 20, Styles.FULL_WIDTH, context));
            setArchetypeNodes(new ArchetypeNodes(getArchetypeNodes()).exclude(RESULT));
        }
        Property notes = properties.get(NOTES);
        if (!StringUtils.isEmpty(notes.getString())) {
            // will display as a single line if short
            addComponent(createMultiLineText(notes, 1, 20, Styles.FULL_WIDTH, context));
        }
        return super.apply(object, properties, parent, context);
    }

    /**
     * Lays out child components in a grid.
     *
     * @param object     the object to lay out
     * @param parent     the parent object. May be {@code null}
     * @param properties the properties
     * @param container  the container to use
     * @param context    the layout context
     */
    @Override
    protected void doSimpleLayout(IMObject object, IMObject parent, List<Property> properties, Component container,
                                  LayoutContext context) {
        ComponentState result = getComponent(RESULT);
        ComponentState notes = getComponent(NOTES);
        ComponentGrid grid = createGrid(object, properties, context, 2);
        addAuditInfo(object, grid, context);
        if (result != null) {
            grid.add(result, 2);
        }
        if (notes != null) {
            grid.add(notes, 2);
        }
        Grid component = createGrid(grid);
        component.setWidth(Styles.FULL_WIDTH);
        container.add(ColumnFactory.create(Styles.INSET, component));
    }
}
