/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.workspace.workflow.messaging.sms;

import nextapp.echo2.app.Component;
import org.apache.commons.lang3.StringUtils;
import org.openvpms.component.business.domain.im.common.IMObject;
import org.openvpms.component.model.bean.IMObjectBean;
import org.openvpms.web.component.im.layout.ArchetypeNodes;
import org.openvpms.web.component.im.layout.LayoutContext;
import org.openvpms.web.component.im.view.ComponentState;
import org.openvpms.web.component.property.Property;
import org.openvpms.web.component.property.PropertySet;
import org.openvpms.web.echo.factory.RowFactory;
import org.openvpms.web.workspace.customer.communication.AbstractCommunicationLayoutStrategy;

import java.util.List;
import java.util.Objects;

import static org.openvpms.web.component.im.layout.ArchetypeNodes.find;
import static org.openvpms.web.component.im.layout.ArchetypeNodes.include;
import static org.openvpms.web.echo.style.Styles.CELL_SPACING;

/**
 * Layout strategy for <em>act.smsMessage</em>.
 *
 * @author Tim Anderson
 */
public class SMSMessageLayoutStrategy extends AbstractCommunicationLayoutStrategy {

    /**
     * The recipient node.
     */
    private static final String RECIPIENT = "recipient";

    /**
     * The phone node.
     */
    private static final String PHONE = "phone";

    /**
     * The customer node.
     */
    private static final String CUSTOMER = "customer";

    /**
     * The status node.
     */
    private static final String STATUS = "status";

    /**
     * The status message node.
     */
    private static final String STATUS_MESSAGE = "statusMessage";

    /**
     * The expiry time node.
     */
    private static final String EXPIRY_TIME = "expiryTime";

    /**
     * The replies node.
     */
    private static final String REPLIES = "replies";


    /**
     * Constructs an {@link SMSMessageLayoutStrategy}.
     */
    public SMSMessageLayoutStrategy() {
        this(true);
    }

    /**
     * Constructs an {@link SMSMessageLayoutStrategy}.
     *
     * @param showPatient if {@code true}, show the patient
     */
    public SMSMessageLayoutStrategy(boolean showPatient) {
        super(showPatient, null, "createdTime");
        ArchetypeNodes nodes = ArchetypeNodes.all()
                .exclude(RECIPIENT, STATUS_MESSAGE)
                .excludeIfEmpty(EXPIRY_TIME, REPLIES);
        setArchetypeNodes(nodes);
    }

    /**
     * Apply the layout strategy.
     *
     * @param object     the object to apply
     * @param properties the object's properties
     * @param parent     the parent object. May be {@code null}
     * @param context    the layout context
     * @return the component containing the rendered {@code object}
     */
    @Override
    public ComponentState apply(IMObject object, PropertySet properties, IMObject parent, LayoutContext context) {
        // render the recipient and phone together
        Property recipient = properties.get(RECIPIENT);
        Property phone = properties.get(PHONE);
        if (!context.isEdit()) {
            IMObjectBean bean = getBean(object);
            if (Objects.equals(bean.getTargetRef(RECIPIENT), bean.getTargetRef(CUSTOMER))) {
                // don't re-display the customer
                getArchetypeNodes().exclude(CUSTOMER);
            }
        }
        Component partyComponent = createComponent(recipient, object, context).getComponent();
        Component phoneComponent = createComponent(phone, object, context).getComponent();
        addComponent(new ComponentState(RowFactory.create(CELL_SPACING, partyComponent, phoneComponent), phone));

        // render the status and status message together, if there is a status message
        Property statusMessage = properties.get(STATUS_MESSAGE);
        if (!StringUtils.isEmpty(statusMessage.getString())) {
            addComponent(createComponentPair(STATUS, STATUS_MESSAGE, object, properties, context));
        }
        return super.apply(object, properties, parent, context);
    }

    /**
     * Returns the properties to display in the header.
     * <p>
     * Note; this does not exclude empty properties the returned properties are used to determine the other nodes
     * to display. Empty properties are excluded via {@link #excludeEmptyHeaderProperties(List)} instead.
     *
     * @param properties the properties
     * @return the header properties
     */
    @Override
    protected List<Property> getHeaderProperties(List<Property> properties) {
        return include(properties, PHONE, DESCRIPTION);
    }

    /**
     * Returns the patient-related properties.
     * <p/>
     * This implementation includes the location and patient
     *
     * @param properties the properties
     * @return the header properties
     */
    @Override
    protected List<Property> getPatientProperties(List<Property> properties) {
        return include(properties, LOCATION, CUSTOMER, PATIENT);
    }

    /**
     * Returns the text properties.
     * <p>
     * These are rendered under each other.
     * <p/>
     * If the statusMessage is present, this implementation includes the status and status message to ensure that
     * they don't take up too much horizontal space.
     *
     * @param properties the properties
     * @param message    the message property
     * @return the text properties
     */
    @Override
    protected List<Property> getTextProperties(List<Property> properties, Property message) {
        List<Property> result = super.getTextProperties(properties, message);
        if (getComponent(STATUS) != null) {
            // if the status and statusMessage have been rendered to together, include them
            Property status = find(properties, STATUS);
            if (status != null) {
                result.add(0, status);
            }
        }
        return result;
    }

    /**
     * Excludes empty fields.
     *
     * @param properties the fields
     * @return the properties to render
     */
    @Override
    protected List<Property> excludeEmptyFields(List<Property> properties) {
        return excludeIfEmpty(properties, REASON);
    }
}
